#!/usr/bin/perl -w
#
#  $Id: parseHmmer3Alignments.pl,v 1.3 2010/04/21 23:00:54 kkeller Exp $
#  fastHmm/fastBlast Alignment Tools
#  http://microbesonline.org/fasthmm (fasthmm@microbesonline.org)
#
#  Parse hmmsearch output into tab-delimited domains including hit coords
#
#  Copyright (C) 2007 The Regents of the University of California
#  All rights reserved.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#  Disclaimer
#
#  NEITHER THE UNITED STATES NOR THE UNITED STATES DEPARTMENT OF ENERGY,
#  NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED,
#  OR ASSUMES ANY LEGAL LIABILITY OR RESPONSIBILITY FOR THE ACCURACY,
#  COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
#  OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
#  PRIVATELY OWNED RIGHTS.
#

# for hmmscan, leave HmmName out
# for hmmsearch, use HmmName to specify

use strict;

die "Usage: ParseHmmAlignments.pl [HmmName] < hmmsearch >> domainsfile\n"
    . "   Returns domain, sequence id, sequence begin to end, domain to end, score, E-value,\n"
    . "   and the list of matching positions in the sequence, as e.g. 5:48,50:100,110:200,\n"
    . "   and then similarly the list of matching positions in the model\n"
    if @ARGV >1;

my $hmmName=shift @ARGV;

my ($ACC,$SEQID,$SEQBEG,$SEQEND,$DOMBEG,$DOMEND,$SCORE,$E) = (0..7);

while(<STDIN>) {
#    last if m/^Parsed for domains/;
	last unless m/^#/;
}

# this starts the hits section
# for hmmscan hits are sorted by query

while (1)
{
	my $done=process_query();
	last if $done;
}

sub process_query {

return 1 if eof;

my $query;

while(<STDIN>) {
	next if /^$/;
    chomp;
	if (/^Query:/)
	{
		($query)=/^Query: +(\w+) +/;
		next;
	}
	
	last if /^Domain annotation/;
}

while (1)
{
	my $done=process_hits($query);
#	warn "done processing $query";
	last if $done;
}


return undef;

} # end sub process_query


sub process_hits {

my $query=shift;
my $acc;
# these hits are for one query to one domain
# there is a summary section before the actual alignments
my @hits = ();
while (<STDIN>)
{
#    next if $_ eq "" || m/^Sequence +Domain/ || m/^[ -]+$/;
	last if /^$/;
	if (/^>>/)
	{
		(undef,$acc,undef)=split /[\t ]+/;
		next;
	}
	next if (/^ +#/ or /^ +-/ or /^ +\[No individual/);
	chomp;
    my @F = split /[\t ]+/, $_;
    die "Cannot parse $_" unless @F==17;
    #warn "$acc $query @F";
    push @hits, [$acc,$query,$F[10],$F[11],$F[7],$F[8],$F[3],$F[6]];
}

while (<STDIN>)
{
	last if (/^ +Alignments for each domain/);
	return 1 if (/^\/\/$/);
}

# we should now be at the first alignment
# using --notextw, we can just take the next seven lines
# the eighth line will either be a new alignment, a new domain,
# or the internal stats (which signals the end of the hits for the
# current gene)

my @lines;
while (1)
{
	my $line=<STDIN>;
	chomp $line;

	push @lines,$line;
	last if $line=~/^$/;
}

OUTER: while (1)
{
	my $hit=shift @hits;
	# don't really need the return value here
	my $done=process_alignment($hit,\@lines) if (ref $hit);
	my $nextLine=<STDIN>;
	chomp $nextLine;
	if ($nextLine =~ /^>>/)
	{
		@hits=();
		(undef,$acc,undef)=split /[\t ]+/, $nextLine;
		while (<STDIN>)
		{
		#    next if $_ eq "" || m/^Sequence +Domain/ || m/^[ -]+$/;
			last if (/^$/);
			next if (/^ +#/ or /^ +-/ or /^ +\[No individual/);
			chomp;
		    my @F = split /[\t ]+/, $_;
		    die "Cannot parse $_" unless @F==17;
		    push @hits, [$acc,$query,$F[10],$F[11],$F[7],$F[8],$F[3],$F[6]];
		}

		while (<STDIN>)
		{
			last if (/^$/ or /^ +Alignments for each domain/);
		}
		$nextLine=<STDIN>;
		chomp $nextLine;
	}

       	if ($nextLine =~ /^ +== domain/)
	{
		@lines=();
		push @lines,$nextLine;
		while (1)
		{
			my $line=<STDIN>;
			chomp $line;
			push @lines,$line;
			last if $line=~/^$/;
		}
		next OUTER;
	} elsif ($nextLine =~ /^$/)
	{
		while (<STDIN>)
		{
			#	warn "done processing $acc for $query";
			return 1 if (/^\/\// or eof);
		}
		die "We shouldn't be here; probable error in parsing";
	}
}

die "We shouldn't be here; probable error in parsing";

} # end sub process_hits


sub process_alignment {

my $hit = shift; # hit object corresponding to alignment we are processing
my $lines = shift;
my @lines=@$lines;

#warn "@lines";
# Lineoffset keeps track of where we are:
# lineoffset = 1 means we expect the model alignment line [or an RF line]
# lineoffset = 2 means we expect the match-quality line (pluses and spaces)
# lineoffset = 3 means we expect the sequence alignment line
# lineoffset = 4 means we expect the blank line that separates the alignment sections
my $lineOffset = 0;
my $modelAlign = undef; # the model alignment
my $seqAlign = undef; # the sequence alignment

# this way, don't even have to loop through!  just parse the needed lines
# hope that the model, plus, and seq lines are always in the same spot
# relative to the end of the hit info
# (because after the domain line, there may or may not be RF or CS lines)
my ($domainLine,$modelAlignLine,$plusLine,$seqAlignLine)=($lines[0],$lines[-5],$lines[-4],$lines[-3]);

my $thisAcc=($hmmName or $hit->[$ACC]);
#warn $thisAcc;
#warn $hit->[$SEQID];
# PANTHER searches include a crazy relative path
die "Cannot parse model line '$modelAlignLine' for model $thisAcc"
	unless (
		$modelAlignLine =~ m/^ +$thisAcc +\d+ +([*<>a-zA-Z.-]+) +\d+ *$/
		or
		$modelAlignLine =~ m/^ +.*books\/$thisAcc.*\/hmmer.hmm +\d+ +([*<>a-zA-Z.-]+) +\d+ *$/
	);
$modelAlign = $1;
#warn $modelAlign;

my ($seqId2,$subStart,$subseq,$subEnd) =
	$seqAlignLine =~ m/^\s*(\S+)\s+([0-9-]+)\s+(\S+)\s+([0-9-]+) ?/;
#warn $seqId2;
$seqAlign=$subseq;
#warn $seqAlign;

$modelAlign =~ s/^[*]-[>]//;
$modelAlign =~ s/[<]-[*]$//;
# The modelAlign and seqAlign should line up
die "Sequence and model for $hit->[$SEQID] beginning at $hit->[$SEQBEG]"
    . " are of different lengths:\n$modelAlign\n$seqAlign\nin ParseHmmer3Alignments.pl"
    if length($modelAlign) != length($seqAlign);

my $seqOff = $hit->[$SEQBEG];
my $modelOff = $hit->[$DOMBEG];

my $range = []; # in sequence (list of begin-to)
my $rangeHMM = [];
for (my $i = 0; $i < length($seqAlign); $i++) {
    my $c = substr($seqAlign,$i,1);
    if (uc($c) eq $c && $c ne "-") { # upper-case characters are matches to the HMM
	if (@$range > 0 && $range->[-1][1] == $seqOff-1) {
	    $range->[-1][1] = $seqOff;
	} else {
	    push @$range, [$seqOff,$seqOff];
	}
	if (@$rangeHMM > 0 && $rangeHMM->[-1][1] == $modelOff-1) {
	    $rangeHMM->[-1][1] = $modelOff;
	} else {
	    push @$rangeHMM, [$modelOff,$modelOff];
	}
    }
    $modelOff++ unless substr($modelAlign,$i,1) eq ".";
    $seqOff++ unless $c eq "-";
}
#print STDERR "Parsing $subStart:$subEnd\nseq $subseq\nmod $modelAlign\nended at seq $seqOff model $modelOff\n";
#print STDERR "Last model range " . join(" ",@{$rangeHMM->[-1]})."\n";

print join("\t",@$hit,
	   join(",", map {$_->[0].":".$_->[1]} @$range),
	   join(",", map {$_->[0].":".$_->[1]} @$rangeHMM))
    ."\n";
return 1;

##########

while(@lines > 0) {
    my $line = shift @lines;

    # example input line: "616154: domain 1 of 1, from 196 to 341: score 10.1, E = 0.0048"
    # First, save existing alignment if we're at the begining of another one, or at the end
    if (($line =~ m/^Histogram/ || $line =~ m/domain.*score/) && defined $hit) {
		$modelAlign =~ s/^[*]-[>]//;
		$modelAlign =~ s/[<]-[*]$//;
	# The modelAlign and seqAlign should line up
	die "Sequence and model for $hit->[$SEQID] beginning at $hit->[$SEQBEG]"
	    . " are of different lengths:\n$modelAlign\n$seqAlign\nin ParseHmmAlignments.pl"
	    if length($modelAlign) != length($seqAlign);

	my $seqOff = $hit->[$SEQBEG];
	my $modelOff = $hit->[$DOMBEG];

	my $range = []; # in sequence (list of begin-to)
	my $rangeHMM = [];
	for (my $i = 0; $i < length($seqAlign); $i++) {
	    my $c = substr($seqAlign,$i,1);
	    if (uc($c) eq $c && $c ne "-") { # upper-case characters are matches to the HMM
		if (@$range > 0 && $range->[-1][1] == $seqOff-1) {
		    $range->[-1][1] = $seqOff;
		} else {
		    push @$range, [$seqOff,$seqOff];
		}
		if (@$rangeHMM > 0 && $rangeHMM->[-1][1] == $modelOff-1) {
		    $rangeHMM->[-1][1] = $modelOff;
		} else {
		    push @$rangeHMM, [$modelOff,$modelOff];
		}
	    }
	    $modelOff++ unless substr($modelAlign,$i,1) eq ".";
	    $seqOff++ unless $c eq "-";
	}
#	print STDERR "Parsing $subStart:$subEnd\nseq $subseq\nmod $modelAlign\nended at seq $seqOff model $modelOff\n";
#	print STDERR "Last model range " . join(" ",@{$rangeHMM->[-1]})."\n";

	print join("\t",@$hit,
		   join(",", map {$_->[0].":".$_->[1]} @$range),
		   join(",", map {$_->[0].":".$_->[1]} @$rangeHMM))
	    ."\n";
    }
    # Then process the line
    if ($line =~ m/^Internal/) {
	last;
    } elsif ($line =~ m/^ +== domain/) {
#	my ($seqId, $begin, $end) = ($1,$2,$3);
#	die "Unexpected domain -- no hits left" if @hits==0;
#	die "Unexpected seqId $seqId $begin $end vs. $hit->[$SEQID] $hit->[$SEQBEG] $hit->[$SEQEND]" unless
#	    $hit->[$SEQID] eq $seqId && $hit->[$SEQBEG] == $begin && $hit->[$SEQEND] == $end;
	my $seqId=$hit->[$SEQID];
	my $begin=$hit->[$SEQBEG];
	my $end=$hit->[$SEQEND];
	$lineOffset = 1;
	$modelAlign = "";
	$seqAlign = "";
    }  elsif (defined $hit && $lineOffset == 1) {
	next if $line eq ""; # There are extra blank lines between alignments and start of histogram section
	my $thisAcc=$hit->[$ACC];
	my $thisSeqId=$hit->[$ACC];
#	warn $thisAcc=$hit->[$ACC];
	die "Cannot parse '$line' for $thisSeqId and $thisAcc" unless $line =~ m/^ +$thisAcc +\d+ +([*<>a-zA-Z.-]+) +\d+ +?$/;
	$modelAlign .= $1;
	# e.g., $modelAlign = "*->LklldkhLipka...ttaeskvFylKmkGDYyRYlaEfatgeerkeaadk<-*"
	# (The ends are optional)
	# Capitalization indicates the strength of the match, not whether or not it is a match,
	# and ... indicates insert states
	$lineOffset = 2;
    } elsif (defined $hit && $lineOffset == 2) {
	$lineOffset = 3; # skip the alignment matches indicator
    } elsif (defined $hit && $lineOffset == 3) {
	# example line: "      616154   196    GRVFRRDF--EDATHAMMFHQVEGLVI--DKGITMASLKGALTEMAR 238 "
	# example line: "        15502     -     -    " #  Hangover line, note absence of right-end count...
	if ($line =~ m/^\s*(\S+)\s+[-]\s+[-]\s+$/) {
	    my $seqId2 = $1;
	    die "Cannot parse $line -- expected sequence $seqId2" unless 
		substr($hit->[$SEQID],0,length($seqId2)) eq $seqId2;
	} elsif ($line =~ m/^\s*(\S+)\s+([0-9-]+)\s+(\S+)\s+([0-9-]+) ?/) {
	    my ($seqId2,$subStart,$subseq,$subEnd) = ($1,$2,$3,$4);
	    # HMMer often truncates the seqid, but it should still match as far as it is present
	    die "Cannot parse $line -- expected sequence $seqId2"
		unless substr($hit->[$SEQID],0,length($seqId2)) eq $seqId2;
	    my $subseqStrip = $subseq;
	    $subseqStrip =~ s/[-]//g;
	    if ($subStart ne "-") {
		die "Cannot parse $line -- subseq $subseq is of wrong length"
		    unless length($subseqStrip) == $subEnd-$subStart+1;
	    }
	    $seqAlign .= $subseq;
	} else {
	    die "Cannot parse $line";
	}
	$lineOffset = 4;
    } elsif ($line eq "" && $lineOffset==4) {
	$lineOffset = 1;
    }
}

return 1;

} # end sub process_alignment
